import io
from typing import List, Tuple
from PIL import Image, ImageFilter, ImageOps, ImageDraw, ImageEnhance
from ..core import ChepyCore, ChepyDecorators


class Multimedia(ChepyCore):
    """The `Multimedia` class is used predominantly to handle image and 
    audio file processing. All the methods within the `Multimedia` class 
    are available in the main **Chepy** class. For coverage, it is important 
    to understand that this class will sometimes coerce non RGBA to RGB and 
    RGB to RGBA formats.
    
    Examples:
        To use the Multimedia class as a standalone, import with
        
        >>> from chepy.modules.multimedia import Multimedia
        >>> m = Multimedia("/path/to/image.png").load_file()
        This will load the image as bytes into Chepy.

        Advanced example using the Multimedia class. We will take 
        our loaded image, convert it split out the RGB color channels, 
        get the blue image, blur it and finally write it to disk. 

        >>> from chepy.modules.multimedia import Multimedia
        >>> m = Multimedia("/path/to/image.png").load_file()
        >>> m.split_color_channels("png")
        {'red': '...', 'blue': '...', 'green': '...'}
        >>> m.get_by_key("blue")
        b"...PNG..."
        >>> m.blur_image("png")
        >>> m.write("/path/to/file.png", as_binary=True)

        This whole operation can be done in one line also.

        >>> from chepy.modules.multimedia import Multimedia
        >>> m = Multimedia("/path/to/image.png").load_file().split_color_channels("png").get_by_key("blue").blur_image().write("/path/to/file.png", as_binary=True)
    """

    def _force_rgba(self, image):  # pragma: no cover
        if image.mode != "RGBA":
            new = image.convert("RGBA")
            return new
        else:
            return image

    def _force_rgb(self, image):  # pragma: no cover
        if image.mode != "RGB":
            new = image.convert("RGB")
            return new
        else:
            return image

    @ChepyDecorators.call_stack
    def resize_image(
        self,
        width: int,
        height: int,
        extension: str = "png",
        resample: str = "nearest",
        quality: int = 100,
    ):
        """Resize an image. 
        
        Args:
            width (int): Required. Width in pixels
            height (int): Required. Height in pixels
            extension (str, optional): File extension of loaded image. Defaults to png
            resample (str, optional): Resample rate. Defaults to "nearest".
            quality (int, optional): Quality of output. Defaults to 100.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("image.png").load_file().resize_image(256, 256, "png")
            >>> c.write_to_file("/path/to/file.png", as_binary=True)
        """
        fh = io.BytesIO()
        if resample == "nearest":
            resample = Image.NEAREST
        elif resample == "antialias":
            resample = Image.ANTIALIAS
        elif resample == "bilinear":
            resample = Image.BILINEAR
        elif resample == "box":
            resample = Image.BOX
        elif resample == "hamming":
            resample = Image.HAMMING
        else:  # pragma: no cover
            raise TypeError(
                "Valid resampling options are: nearest, antialias, bilinear, box and hamming"
            )
        image = Image.open(self._load_as_file())
        resized = image.resize((width, height), resample=resample)
        resized.save(fh, extension, quality=quality)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def split_color_channels(self, extension: str = "png"):
        """Split an image into its red, green and blue channels

        Args:
            extension (str, optional): File extension of loaded image. Defaults to png
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            Write the blue image to disk in this example

            >>> c = Chepy("logo.png").load_file()
            >>> c.split_color_channels("png")
            {'red': b'...', 'green': b'...', 'blue': b'...'}
            >>> c.get_by_key("blue").write("/path/to/file.png", as_binary=True)
        """
        hold = {}
        image = Image.open(self._load_as_file())
        image = self._force_rgba(image)
        data = image.getdata()

        red = []
        green = []
        blue = []
        for d in data:
            red.append((d[0], 0, 0))
            green.append((0, d[1], 0))
            blue.append((0, 0, d[2]))

        red_fh = io.BytesIO()
        image.putdata(red)
        image.save(red_fh, extension)
        hold["red"] = red_fh.getvalue()

        green_fh = io.BytesIO()
        image.putdata(green)
        image.save(green_fh, extension)
        hold["green"] = green_fh.getvalue()

        blue_fh = io.BytesIO()
        image.putdata(blue)
        image.save(blue_fh, extension)
        hold["blue"] = blue_fh.getvalue()

        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def rotate_image(self, rotate_by: int, extension: str = "png"):
        """Rotate an image
        
        Args:
            rotate_by (int): Required. Roate by degrees
            extension (str, optional): File extension of loaded image. Defaults to png
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            To flip an image horizontally, we can:

            >>> c = Chepy("logo.png").load_file().rotate_image(180, "png")
            >>> c.write('/path/to/file.png', as_binary=True)
        """
        image = Image.open(self._load_as_file())
        fh = io.BytesIO()
        rotated = image.rotate(rotate_by)
        rotated.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def blur_image(
        self, extension: str = "png", gaussian: bool = False, radius: int = 2
    ):
        """Blur an image
        
        Args:
            extension (str, optional): File extension of loaded image. Defaults to png
            gaussian (bool, optional): If Gaussian blur is to be applied. Defaults to False.
            radius (int, optional): Radius for Gaussian blur. Defaults to 2.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("logo.png").load_file().blur_image("png")
            >>> >>> c.write('/path/to/file.png', as_binary=True)

            To apply Gaussian blur, use:

            >>> c = Chepy("logo.png").load_file()
            >>> c.blur_image(extension="png", gaussian=True, radius=4)
            >>> >>> c.write('/path/to/file.png', as_binary=True)
        """
        image = Image.open(self._load_as_file())
        fh = io.BytesIO()
        if gaussian:
            blurred = image.filter(ImageFilter.GaussianBlur(radius=radius))
        else:
            blurred = image.filter(ImageFilter.BLUR)
        blurred.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def grayscale_image(self, extension: str = "png"):
        """Grayscale an image
        
        Args:
            extension (str, optional): File extension of loaded image. Defaults to png
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("logo.png").load_file().grayscale_image("png")
            >>> >>> c.write('/path/to/file.png', as_binary=True)
        """
        image = Image.open(self._load_as_file())
        fh = io.BytesIO()
        gray = image.convert("LA")
        gray.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def invert_image(self, extension: str = "png"):
        """Invert the colors of the image
        
        Args:
            extension (str, optional): File extension of loaded image. Defaults to png
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("logo.png").load_file().invert_image("png")
            >>> >>> c.write('/path/to/file.png', as_binary=True)
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgb(image)
        fh = io.BytesIO()
        inverted = ImageOps.invert(image)
        inverted.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_opacity(self, level: int, extension: str = "png"):
        """Change the opacity of an image
        
        Args:
            level (int): Required. Level of opacity. Half is 128
            extension (str, optional): File extension of loaded image. Defaults to png
        
        Returns:
            Chepy: The Chepy object. 
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgba(image)
        fh = io.BytesIO()
        image.putalpha(level)
        image.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_contrast(self, factor: int, extension: str = "png"):
        """Change image contrast
        
        Args:
            factor (int): Factor to increase the contrast by
            extension (str, optional): File extension of loaded image. Defaults to "png"
        
        Returns:
            Chepy: The Chepy object. 
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgb(image)
        fh = io.BytesIO()
        enhanced = ImageEnhance.Contrast(image).enhance(factor)
        enhanced.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_brightness(self, factor: int, extension: str = "png"):
        """Change image brightness
        
        Args:
            factor (int): Factor to increase the brightness by
            extension (str, optional): File extension of loaded image. Defaults to "png"
        
        Returns:
            Chepy: The Chepy object. 
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgb(image)
        fh = io.BytesIO()
        enhanced = ImageEnhance.Brightness(image).enhance(factor)
        enhanced.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_sharpness(self, factor: int, extension: str = "png"):
        """Change image sharpness
        
        Args:
            factor (int): Factor to increase the sharpness by
            extension (str, optional): File extension of loaded image. Defaults to "png"
        
        Returns:
            Chepy: The Chepy object. 
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgb(image)
        fh = io.BytesIO()
        enhanced = ImageEnhance.Sharpness(image).enhance(factor)
        enhanced.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_color(self, factor: int, extension: str = "png"):
        """Change image color
        
        Args:
            factor (int): Factor to increase the color by
            extension (str, optional): File extension of loaded image. Defaults to "png"
        
        Returns:
            Chepy: The Chepy object. 
        """
        image = Image.open(self._load_as_file())
        image = self._force_rgb(image)
        fh = io.BytesIO()
        enhanced = ImageEnhance.Color(image).enhance(factor)
        enhanced.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_to_asciiart(
        self,
        art_width: int = 120,
        chars: List[str] = ["B", "S", "#", "&", "@", "$", "%", "*", "!", ":", "."],
    ):  # pragma: no cover
        """Convert image to ascii art

        `Reference: <https://dev.to/anuragrana/generating-ascii-art-from-colored-image-using-python-4ace>`__
        
        Args:
            art_width (int, optional): Width of the ascii art. Higher values 
                show more details. Defaults to 120.
            chars (List[str], optional): A list of chars to build the ascii art with
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("pythonlogo.png").load_file()
            >>> c.image_to_asciiart().o
            .::.:..................................................:.::.
            .::......................................................::.
            :.......................&SSSSSSSSSS!.......................:
            ...................*SSSSSSSSSSSSSSSSSSSS....................
            ..................SSSS$@SSSSSSSSSSSSSSSSSS..................
            .................:SSS....SSSSSSSSSSSSSSSSSS.................
            .................:SSS....SSSSSSSSSSSSSSSSSS.................
            .................:SSSSSSSSSSSSSSSSSSSSSSSSS.................
            ..................SSSSSSSSSSSSSSSSSSSSSSSSS.................
            ..............................SSSSSSSSSSSSS.*****!..........
            ........SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS.*********.......
            ......SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS.**********......
            .....SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS.***********.....
            .....SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS..************....
            ....SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS..************....
            ....SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS...*************....
            ....SSSSSSSSSSSSSSSS.....................***************....
            ....SSSSSSSSSSSSS....***********************************....
            ....SSSSSSSSSSSS...*************************************....
            .....SSSSSSSSSSS..**************************************....
            .....SSSSSSSSSSS..*************************************.....
            ......SSSSSSSSSS..************************************......
            .......SSSSSSSSS..**********************************!.......
            ..........SSSSSS..************..............................
            ..................*************************.................
            ..................*************************.................
            ..................******************...****.................
            ..................*****************.....***.................
            ..................******************:.****..................
            ....................*********************...................
            :.......................!***********:......................:
            .::......................................................::.
            .::.:..................................................:.::.
        """
        img = Image.open(self._load_as_file())

        width, height = img.size
        aspect_ratio = height / width
        new_width = art_width
        new_height = aspect_ratio * new_width * 0.55
        img = img.resize((new_width, int(new_height)))
        img = img.convert("L")

        pixels = img.getdata()

        new_pixels = [chars[pixel // 25] for pixel in pixels]
        new_pixels = "".join(new_pixels)

        new_pixels_count = len(new_pixels)
        ascii_image = [
            new_pixels[index : index + new_width]
            for index in range(0, new_pixels_count, new_width)
        ]
        self.state = "\n".join(ascii_image)
        return self

    @ChepyDecorators.call_stack
    def convert_image(self, format_to: str):
        """Change image format. 

        Example, convert png to jpeg
        
        Args:
            format_to (str): Required. A valid image format extension
        
        Returns:
            Chepy: The Chepy object

        Examples:
            For example, to change a png image to a jpeg image, and save it:

            >>> c = Chepy("logo.png").load_file().convert_image("jpeg")
            b'\\xff\\xd8\\xff\\xe0...'
            >>> c.write("/path/to/file.jpeg", as_binary=True)
        """
        image = Image.open(self._load_as_file())
        fh = io.BytesIO()

        if image.mode != "RGB":
            new = image.convert("RGB")
        else:  # pragma: no cover
            new = image

        new.save(fh, format_to)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def image_add_text(
        self,
        text: str,
        extension: str = "png",
        coordinates: Tuple[int, int] = (0, 0),
        color: Tuple[int, int, int] = (0, 0, 0),
    ):
        """Add text to an image
        
        Args:
            text (str): Required. Text to add
            extension (str, optional): File extension of loaded image. Defaults to png
            coordinates (Tuple[int, int], optional): Coordinates of image where to add text. 
                Defaults to (0, 0).
            color (Tuple[int, int, int], optional): Color of text. 
                Defaults to (0, 0, 0) which is black.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("logo.png").load_file().image_add_text("some text")
            b'\\xff\\xd8\\xff\\xe0...'
            >>> c.write("/path/to/file.jpeg", as_binary=True)
        """
        image = Image.open(self._load_as_file())
        fh = io.BytesIO()

        if image.mode != "RGB":
            new = image.convert("RGB")
        else:  # pragma: no cover
            new = image

        ImageDraw.Draw(new).text(coordinates, text, color)
        new.save(fh, extension)
        self.state = fh.getvalue()
        return self

    @ChepyDecorators.call_stack
    def lsb_dump_by_channel(self, channel: str = "r", column_first: bool = False): # pragma: no cover
        """Dump LSB from a specific color channel
        
        Args:
            channel (str, optional): Color channel. r, g, b. Defaults to 'r'.
            column_first (bool, optional): Order by column first. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 
        """
        channels = ["r", "g", "b"]
        assert channel in channels, "Valid channels are {}".format("".join(channels))
        if channel == "r":
            index = 0
        elif channel == "g":
            index = 1
        elif channel == "b":
            index = 2
        image = Image.open(self._load_as_file())
        pixels = image.load()
        if column_first:
            height, width = image.size
        else:
            width, height = image.size

        data = ""
        for y in range(height):
            for x in range(width):
                try:
                    pix = pixels[x, y][index]
                    lsb = bin(pix).zfill(8)[-1]
                    data += lsb
                except IndexError:
                    break

        self.state = data
        return self

    @ChepyDecorators.call_stack
    def msb_dump_by_channel(self, channel: str = "r", column_first: bool = False): # pragma: no cover
        """Dump MSB from a specific color channel
        
        Args:
            channel (str, optional): Color channel. r, g, b. Defaults to 'r'.
            column_first (bool, optional): Order by column first. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 
        """
        channels = ["r", "g", "b"]
        assert channel in channels, "Valid channels are {}".format("".join(channels))
        if channel == "r":
            index = 0
        elif channel == "g":
            index = 1
        elif channel == "b":
            index = 2
        image = Image.open(self._load_as_file())
        pixels = image.load()
        if column_first:
            height, width = image.size
        else:
            width, height = image.size

        data = ""
        for x in range(height):
            for y in range(width):
                val = pixels[x, y][index]
                data += ((bin(val)[2:]).zfill(8))[0]

        self.state = data
        return self



import collections
import socket
import ssl
import ipaddress
import regex as re
import urllib.parse as _py_urlparse
from ..core import ChepyCore, ChepyDecorators


class Networking(ChepyCore):
    @ChepyDecorators.call_stack
    def defang_url(self):
        """Make a URL harmless
        
        Takes a Universal Resource Locator (URL) and 'Defangs' it; 
        meaning the URL becomes invalid, neutralising the risk of accidentally 
        clicking on a malicious link. This is often used when dealing with 
        malicious links or IOCs.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("https://app.google.com/?lol=some data&a=1").defang_url().o
            "hxxps://app[.]google[.]com/?lol=some data&a=1"
        """
        self.state = re.sub(r"(^htt)", "hxx", self._convert_to_str())
        self.state = re.sub(r"\.", "[.]", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def refang_url(self):
        """Refangs a URL so that it is clickable
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("hxxps://app[.]google[.]com/?lol=some data&a=1").refang_url().o
            "https://app.google.com/?lol=some data&a=1"
        """
        self.state = re.sub(r"(^hxx)", "htt", self._convert_to_str())
        self.state = re.sub(r"\[\.\]", ".", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def defang_ip(self):
        """Make an IP address harmless
        
        Takes a IPv4 or IPv6 address and 'Defangs' it, meaning the 
        IP becomes invalid, removing the risk of accidentally utilising 
        it as an IP address.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("2001:4860:4860::8844").defang_ip().o
            "2001[:]4860[:]4860[:][:]8844"
            
            >>> Chepy("127.0.0.1").defang_ip().o
            "127[.]0[.]0[.]1"
        """
        if ":" in self._convert_to_str():
            self.state = re.sub(r":", "[:]", self._convert_to_str())
        else:
            self.state = re.sub(r"\.|:", "[.]", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def refang_ip(self):
        """Refangs an IP address
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("127[.]0[.]0[.]1").refang_ip().o
            "127.0.0.1"
        """
        self.state = re.sub(r"\[\.\]|\[\:\]", ".", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def parse_uri(self):
        """Parse a URI
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("http://example.com/resource?foo=bar#fragment").parse_uri().o
            {
                "scheme": "http",
                "location": "example.com",
                "path": "/resource",
                "params": "",
                "query": {"foo": ["bar"]},
                "fragment": "fragment",
            }
        """
        parsed = _py_urlparse.urlparse(self._convert_to_str())
        self.state = {
            "scheme": parsed.scheme,
            "location": parsed.netloc,
            "path": parsed.path,
            "params": parsed.params,
            "query": _py_urlparse.parse_qs(parsed.query),
            "fragment": parsed.fragment,
        }
        return self

    @ChepyDecorators.call_stack
    def parse_ip_range(self):
        """Enumerate IP address in a CIDR range
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("10.10.10.1/24").parse_ip_range().o
            [
                "10.10.10.1",
                "10.10.10.2,
                ...
                "10.10.10.254"
            ]
        """
        self.state = [
            str(i)
            for i in ipaddress.ip_network(self._convert_to_str(), strict=False).hosts()
        ]
        return self

    @ChepyDecorators.call_stack
    def parse_ipv6(self):
        """Get longhand and shorthand of IPv6
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("2001:4860:4860::8888").parse_ipv6().o
            {
                "long": "2001:4860:4860:0000:0000:0000:0000:8888",
                "short": "2001:4860:4860::8888",
            }
        """
        ip = ipaddress.ip_address(self._convert_to_str())
        self.state = {"long": ip.exploded, "short": ip.compressed}
        return self

    @ChepyDecorators.call_stack
    def get_ssl_cert(self, port: int = 443):
        """Get the server side SSL certificate for a domain
        
        Args:
            port (int, optional): Server port. Defaults to 443.
        
        Returns:
            Chepy: The Chepy object

        Examples:
            >>> Chepy('google.com').get_ssl_cert().o
            {
                'subject': {
                    'commonName': '*.google.com',
                    'organizationName': 'Google LLC',
                ...
                'caIssuers': ('http://pki.goog/gsr2/GTS1O1.crt',),
                'crlDistributionPoints': ('http://crl.pki.goog/GTS1O1.crl',)
            }
        """
        domain = re.sub("^\w+://", "", self._convert_to_str())
        with socket.create_connection((domain, port)) as sock:
            context = ssl.create_default_context()
            with context.wrap_socket(sock, server_hostname=domain) as sslsock:
                cert = sslsock.getpeercert()
                final = {}
                for key in cert.keys():
                    if key == "subject" or key == "issuer":
                        final[key] = dict(
                            collections.ChainMap(*list(map(dict, cert[key])))
                        )
                    elif key == "subjectAltName":
                        final[key] = list(
                            map(lambda x: dict([x]), cert["subjectAltName"])
                        )
                    else:
                        final[key] = cert[key]
                self.state = final
                return self



import ujson
import pydash
import phpserialize
import regex as re
from lxml import etree
from ..core import ChepyCore, ChepyDecorators


class CodeTidy(ChepyCore):
    @ChepyDecorators.call_stack
    def minify_json(self):
        """Minify JSON string
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("/path/to/file.json").load_file()
            >>> print(c.minify_json())
        """
        self.state = ujson.dumps(ujson.loads(self._convert_to_str()))
        return self

    @ChepyDecorators.call_stack
    def beautify_json(self, indent: int = 2):
        """Beautify minified JSON
        
        Args:
            indent (int, optional): Indent level. Defaults to 2.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("/path/to/file.json").load_file()
            >>> print(c.beautify_json(indent=4))
        """
        self.state = ujson.dumps(ujson.loads(self._convert_to_str()), indent=indent)
        return self

    @ChepyDecorators.call_stack
    def minify_xml(self):
        """Minify XML string
        
        Returns:
            Chepy: The Chepy object. 
        
        Examples:
            >>> c = Chepy("/path/to/file.xml").load_file()
            >>> print(c.minify_xml())
        """
        parser = etree.XMLParser(remove_blank_text=True)
        self.state = etree.tostring(
            etree.fromstring(self._convert_to_bytes(), parser=parser)
        )
        return self

    @ChepyDecorators.call_stack
    def beautify_xml(self):
        """Beautify compressed XML
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("/path/to/file.xml").load_file()
            >>> print(c.beautify_json())
        """
        self.state = etree.tostring(
            etree.fromstring(self._convert_to_bytes()), pretty_print=True
        )
        return self

    @ChepyDecorators.call_stack
    def php_deserialize(self):
        """Deserialize php to dict
        
        Deserializes PHP serialized data, outputting keyed arrays as a python dict.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy('a:3:{i:1;s:6:"elem 1";i:2;s:6:"elem 2";i:3;s:7:" elem 3";}')
            >>> c.php_deserialize()
            {1: b'elem 1', 2: b'elem 2', 3: b' elem 3'}
        """
        self.state = phpserialize.loads(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def to_upper_case(self, by: str = "all"):
        """Convert string to uppercase
        
        Args:
            by (str, optional): Convert all, by word or by sentence. Defaults to 'all'.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            Uppercase by word

            >>> Chepy("some String").to_upper_case(by="word").o
            "Some String"
            
            Uppercase by sentence

            >>> Chepy("some String").to_upper_case(by="sentence").o
            "Some string"
            
            Uppercase all

            >>> Chepy("some String").to_upper_case(by="all").o
            "SOME STRING"
        """
        assert by in [
            "all",
            "word",
            "sentence",
        ], "Valid options are all, word and sentence"
        if by == "all":
            self.state = self._convert_to_str().upper()
        elif by == "word":
            self.state = self._convert_to_str().title()
        elif by == "sentence":
            self.state = self._convert_to_str().capitalize()
        return self

    @ChepyDecorators.call_stack
    def to_lower_case(self):
        """Convert string to lowercase

        Converts every character in the input to lower case.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("HelLo WorLd").to_lower_case().o
            "hello world"
        """
        self.state = self._convert_to_str().lower()
        return self

    @ChepyDecorators.call_stack
    def to_snake_case(self):
        """Convert string to snake case

        Converts the input string to snake case. Snake case is all lower case 
        with underscores as word boundaries. e.g. this_is_snake_case.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("helloWorld").to_snake_case().o
            "hello_world"
        """
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", self._convert_to_str())
        self.state = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
        return self

    @ChepyDecorators.call_stack
    def to_camel_case(self, ignore_space: bool = False):
        """Convert string to camel case
        
        Converts the input string to camel case. Camel case is all lower case 
        except letters after word boundaries which are uppercase. e.g. thisIsCamelCase 

        Args:
            ignore_space (bool, optional): Ignore space boundaries. Defaults to False.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some Data_test").to_camel_case().o
            "someDataTest"
            
            To ignore space, we can set the `ignore_space` to True
            >>> Chepy("some Data_test").to_camel_case(ignore_space=True).o
            "some DataTest"
        """
        if ignore_space:
            r = re.compile(r"_.|\-.")
        else:
            r = re.compile(r"_.|\-.|\s.")
        self.state = r.sub(lambda x: x.group()[1].upper(), self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_kebab_case(self):
        """Convert string to kebab case

        Converts the input string to kebab case. Kebab case is all lower case 
        with dashes as word boundaries. e.g. this-is-kebab-case.

        Returns:
            Chepy: The Chepy object.

        Examples:  
            >>> Chepy("Some data_test").to_kebab_case().o
            "some-data-test"
        """
        self.state = pydash.kebab_case(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def swap_case(self):
        """Swap case in a string
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("SoMe TeXt").swap_case().o
            "sOmE tExT"
        """
        self.state = pydash.swap_case(self._convert_to_str())
        return self



import logging

try:
    import exiftool
except ImportError:
    logging.warning("Could not import exiftool. Use pip install exiftool")

import chepy.core


class Chepy_Exif(chepy.core.ChepyCore):
    """This plugin allows exif data from files and images
    """

    @chepy.core.ChepyDecorators.call_stack
    def get_exif(self):  # pragma: no cover
        """Extract EXIF data from a file
        
        Returns:
            Chepy: The Chepy object. 
        """
        filename = self._temp_file()
        with exiftool.ExifTool() as et:
            exif = et.get_metadata(filename)
            if exif:
                self.state = exif
            else:
                self._warning_logger("No exif data found")
        return self



import regex as re

from ..core import ChepyCore, ChepyDecorators


class Links(ChepyCore):
    @ChepyDecorators.call_stack
    def pastebin_to_raw(self):
        """Convert a pastebin link to raw pastebin link
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("https://pastebin.com/abCD").pastebin_to_raw()
            'https://pastebin.com/raw/abCD'
        """
        self.state = re.sub(r"(pastebin\.com)(/)", r"\1/raw\2", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def github_to_raw(self):
        """Convert a github link to raw github link
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("https://github.com/securisec/chepy/blob/master/README.md").github_to_raw()
            'https://raw.githubusercontent.com/securisec/chepy/master/README.md'
        """
        self.state = re.sub(
            "/blob",
            "",
            re.sub(
                "(github\.com)(/)",
                r"raw.githubusercontent.com\2",
                self._convert_to_str(),
            ),
        )
        return self



import collections
import math
from typing import Any


def shannon_entropy(data: Any, unit="shannon") -> float:
    """Calculate the Shannon entropy of any data type. Units 
    could be `natural`, `shannon` or `hartley`

    - 0 represents no randomness (i.e. all the bytes in the data have the same value) 
        whereas 8, the maximum, represents a completely random string.
    - Standard English text usually falls somewhere between 3.5 and 5.
    - Properly encrypted or compressed data of a reasonable length should have an entropy of over 7.5.

    Args:
        data (Any): Any data type
        unit (str, optional): Unit type. Defaults to "shannon".
    
    Returns:
        float: Calculated entropy
    """
    base = {"shannon": 2.0, "natural": math.exp(1), "hartley": 10.0}

    if len(data) <= 1:  # pragma: no cover
        return 0

    counts = collections.Counter()

    for d in data:
        counts[d] += 1

    ent = 0

    probs = [float(c) / len(data) for c in counts.values()]
    for p in probs:
        if p > 0.0:
            ent -= p * math.log(p, base[unit])

    return ent


def index_of_coincidence(data: Any) -> float:
    """Index of Coincidence (IC) is the probability of two randomly 
    selected characters being the same.

    - 0 represents complete randomness (all characters are unique), whereas 1 represents no randomness (all characters are identical).
    - English text generally has an IC of between 0.67 to 0.78.
    - 'Random' text is determined by the probability that each letter occurs the same number of times as another.

    `Reference <https://gist.github.com/enigmaticape/4254054>`__
    
    Args:
        data (Any): Any data type
    
    Returns:
        float: IC as float
    """
    cipher_flat = "".join([x.upper() for x in data.split() if x.isalpha()])

    N = len(cipher_flat)
    freqs = collections.Counter(cipher_flat)
    alphabet = map(chr, range(ord("A"), ord("Z") + 1))
    freqsum = 0.0

    for letter in alphabet:
        freqsum += freqs[letter] * (freqs[letter] - 1)

    try:
        IC = freqsum / (N * (N - 1))
    except ZeroDivisionError:
        IC = 0
        
    return IC



import ujson
from typing import Iterator, Dict
from urllib.request import urlopen
from Crypto.PublicKey import RSA

from .combinatons import generate_combo, hex_chars
from chepy import Chepy


def factordb(n: int) -> dict:  # pragma: no cover
    """Query the factordb api and get primes if available
    
    Args:
        n (int): n is the modulus for the public key and the private keys
    
    Returns:
        dict: response from api as a dictionary. None if status code is not 200
    """
    res = urlopen("http://factordb.com/api/?query={}".format(str(n)))
    if res.status != 200:
        return None
    return ujson.loads(res.read().decode())


def construct_private_key(
    n: int, e: int, d: int, format: str = "PEM", passphrase: str = None
) -> str:  # pragma: no cover
    """Construct a private key given n, e and d
    
    Args:
        n (int): n
        e (int): e
        d (int): d
        format (str, optional): Supports PEM, DER and OpenSSH. Defaults to "PEM".
        passphrase (str, optional): [description]. Defaults to None.
    
    Returns:
        str: Private key
    """
    valid_formats = ["PEM", "DER", "OpenSSH"]
    assert format in valid_formats, "Valid formats are {}".format(
        " ".join(valid_formats)
    )
    priv = RSA.construct((n, e, d))
    return priv.export_key(format=format, passphrase=passphrase)


def xor_bruteforce_multi(
    data: str, min: int = 0, max: int = None, errors: str = "backslashreplace"
) -> Iterator[Dict[str, str]]:
    """Bruteforce multibyte xor encryption. For faster results, use pypy3. 
    It is important to set the min and max values if the key size is known.
    
    Args:
        data (str): XORed data
        min (int, optional): Minimum key length. Default will start at 1 byte
            . Defaults to 0.
        max (int, optional): Maximum key length. Maximum value is 257 bytes. Defaults to None.
        errors (str, optional): How should the errors be handled? Defaults to backslashreplace.
            Valid options are replace, ignore, backslashreplace
    
    Returns:
        Iterator[Dict[str, str]]: A dictionary where key is key, and value is xored data
    
    Yields:
        Iterator[Dict[str, str]]: A generator which contains a dictionary with the 
            keys: `key` and `out`
    """
    for key in generate_combo(
        hex_chars(), min_length=min, max_length=max, join_by=""
    ):  # pragma: no cover
        yield {
            "key": key,
            "out": Chepy(data).xor(key).bytearray_to_str(errors=errors).o,
        }



import difflib
import regex as re
import exrex
from collections import OrderedDict
from typing import Any, Union
import pydash

from ..core import ChepyCore, ChepyDecorators
from .exceptions import StateNotDict, StateNotList
import chepy.modules.internal.colors as _int_colors


class Utils(ChepyCore):
    @ChepyDecorators.call_stack
    def reverse(self, count: int = 1):
        """Reverses a string
        
        Args:
            count (int, optional): Reverse by the number of characters indicated in count. Defaults to 1.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abcdefg").reverse().output
            "gfedcba"
        """
        if count == 1:
            self.state = self.state[::-1]
            return self
        else:
            self.state = "".join(
                [self.state[x : x + count] for x in range(0, len(self.state), count)][
                    ::-1
                ]
            )
            return self

    @ChepyDecorators.call_stack
    def count_occurances(self, regex: str, case_sensitive: bool = False):
        """Counts occurances of the regex.

        Counts the number of times the provided string occurs.

        Args:
            regex (str): Required. Regex string to search for
            case_sensitive (bool, optional): If search should be case insensitive, by default False

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("AABCDADJAKDJHKSDAJSDdaskjdhaskdjhasdkja").count_occurances("ja").output
            2
        """
        if case_sensitive:
            r = re.compile(regex)
        else:
            r = re.compile(regex, re.IGNORECASE)
        self.state = len(r.findall(self._convert_to_str()))
        return self

    @ChepyDecorators.call_stack
    def remove_whitespace(
        self,
        spaces: bool = True,
        carriage_return: bool = True,
        line_feeds: bool = True,
        tabs: bool = True,
        form_feeds: bool = True,
    ):
        """Remove whitespace from a string
        
        Args:
            spaces (bool, optional): Remove spaces. Defaults to True.
            carriage_return (bool, optional): Remove carriage return \\r. Defaults to True.
            line_feeds (bool, optional): Remove line feeds \\n. Defaults to True.
            tabs (bool, optional): Temove tabs \\t. Defaults to True.
            form_feeds (bool, optional): Remove form feeds \\f. Defaults to True.
        
        Returns:
            Chepy: The Chepy object.
        """
        remove = []
        if spaces:
            remove.append(" ")
        if carriage_return:
            remove.append("\r")
        if line_feeds:
            remove.append("\n")
        if tabs:
            remove.append("\t")
        if form_feeds:
            remove.append("\f")
        self.state = re.sub("|".join(remove), "", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def remove_nullbytes(self):
        """Remove null \\x00 byes from binary data
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().replace(b"\x00", b"")
        return self

    @ChepyDecorators.call_stack
    def regex_search(
        self,
        pattern: str,
        ignore_case: bool = False,
        multiline: bool = False,
        dotall: bool = False,
        unicode: bool = False,
        extended: bool = False,
    ):
        """Regex search on current data
        
        Args:
            pattern (str): Required. The regex pattern to search by
            ignore_case (bool, optional): Set case insentive flag. Defaults to False.
            multiline (bool, optional): ^/$ match start/end. Defaults to False.
            dotall (bool, optional): `.` matches newline. Defaults to False.
            unicode (bool, optional): Match unicode characters. Defaults to False.
            extended (bool, optional): Ignore whitespace. Defaults to False.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("loLolololoL")
            >>> c.regex_search("ol", ignore_case=True)
        """
        flags = 0
        if ignore_case:
            flags += re.IGNORECASE
        if multiline:
            flags += re.MULTILINE
        if dotall:
            flags += re.DOTALL
        if unicode:
            flags += re.UNICODE
        if extended:
            flags += re.X
        self.state = re.findall(pattern, self._convert_to_str(), flags=flags)
        return self

    @ChepyDecorators.call_stack
    def split_by(self, pattern: str = "\n", trim=True):
        """Split a string by the given pattern
        
        Args:
            pattern (str, optional): Pattern to split by. Defaults to '\\n'.
            time (bool, optional): Trim whitespace after split. Defaults to True
        
        Returns:
            Chepy: The Chepy object.
        """
        if trim:
            self.state = list(
                map(pydash.trim, re.split(pattern, self._convert_to_str()))
            )
        else:
            self.state = re.split(pattern, self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def split_by_n(self, n: int):
        """Split a string by n characters.
        
        Args:
            n (int): n from 0
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some string").split_by_n(2).o[2]
            " s"
        """
        self.state = re.findall(".{1," + str(n) + "}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def select_every_n(self, n: int, start: int = 0):
        """Select every nth item from a list or string. 

        Index starts at 0
        
        Args:
            n (int): n from 0
            start (int): starting position. Defaults to 0.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy(["a", 1, "lol", "b", True]).select_every_n(3)
            ["a", "b"]
        """
        self.state = self.state[start::n]
        return self

    @ChepyDecorators.call_stack
    def unique(self):
        """Get an array of unique list items
        
        Raises:
            StateNotList: If state is not a list
        
        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        self.state = pydash.uniq(self.state)
        return self

    @ChepyDecorators.call_stack
    def sort_list(self, reverse: bool = False):
        """Sort a list
        
        Args:
            reverse (bool, optional): In reverse order. Defaults to False.
        
        Raises:
            StateNotList: If state is not list
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(["a", "b", "1", "2"]).sort_list().o
            ["1", "2", "a", "b"]
        """
        assert isinstance(self.state, list), StateNotList()
        self.state = sorted(
            self.state, key=lambda v: (isinstance(v, str), v), reverse=reverse
        )
        return self

    @ChepyDecorators.call_stack
    def sort_dict_key(self, reverse: bool = False):
        """Sort a dictionary by key
        
        Args:
            reverse (bool, optional): Reverse sort order. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy({'z': 'string', 'a': True, 'zz': 1, 'aaa': {'bb': 'data'}, 'ccc': [1,'a']})
            >>> c.sort_dict_key(reverse=True)
            {'zz': 1, 'z': 'string', 'ccc': [1, 'a'], 'aaa': {'bb': 'data'}, 'a': True}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = dict(OrderedDict(sorted(self.state.items(), reverse=reverse)))
        return self

    @ChepyDecorators.call_stack
    def sort_dict_value(self, reverse=False):
        """Sort dictionary by value
        
        Args:
            reverse (bool, optional): Reverse sort order. Defaults to False.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy({'z': 'string', 'a': 'True', 'zz': '1', 'aaa': {'bb': 'data'}, 'ccc': [1,'a']})
            >>> c.sort_dict_value()
            {'zz': '1', 'a': 'True', 'ccc': [1, 'a'], 'z': 'string', 'aaa': {'bb': 'data'}}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = dict(
            OrderedDict(
                sorted(self.state.items(), reverse=reverse, key=lambda x: str(x[1]))
            )
        )
        return self

    @ChepyDecorators.call_stack
    def filter_list(self, by: Union[str, dict], regex: bool = True):
        """Filter a list by a string regex or dict key
        
        Args:
            by (Union[str, dict]): If string, supports regex. Or dictionary
            regex (bool, optional): If pattern is a regex. Defaults to True
        
        Raises:
            StateNotList: If state is not a list
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy('[{"a": 1}, {"b": 2}, {"a": 1, "b": 3}]').str_list_to_list().filter_list("b").o
            [{"b": 2}, {"a": 1, "b": 3}]
        """
        assert isinstance(self.state, list), StateNotList()
        if regex:
            pattern = by if isinstance(self.state[0], str) else by.encode()
            self.state = [f for f in self.state if re.search(pattern, f)]
        else:
            self.state = pydash.filter_(self.state, by)
        if len(self.state) == 1:
            self.state = self.state[0]
        return self

    @ChepyDecorators.call_stack
    def filter_list_by_length(self, length: int, exact: bool = False):
        """Filter a list by length by specifying minimum length. 

        It will also return items that exceed the specified length. 
        
        Args:
            length (int): Minimum length to match
            exact (bool): Match exact length
        
        Returns:
            Chepy: The Chepy object. 
        """
        assert isinstance(self.state, list), StateNotList()
        if exact:
            self.state = [x for x in self.state if len(str(x)) == int(length)]
        else:
            self.state = [x for x in self.state if len(str(x)) >= int(length)]
        return self

    @ChepyDecorators.call_stack
    def filter_dict_key(self, by: str):
        """Filter dictionary by key
        
        Args:
            by (str): Required. Key to filter by. 
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy({'some': 'dict', 'another': 'val'}).filter_dict_key('ano')
            {'another': 'val'}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = {
            key: val for (key, val) in self.state.items() if re.search(by, str(key))
        }
        return self

    @ChepyDecorators.call_stack
    def filter_dict_value(self, by: str):
        """Filter dictionary by value. 

        This method does descend into nested dictionary values. 
        
        Args:
            by (str): Required. Value to filter by. 
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy({'some': 'dict', 'another': 'val'}).filter_dict_value('val')
            {'another': 'val'}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = {
            key: val for (key, val) in self.state.items() if re.search(by, str(val))
        }
        return self

    @ChepyDecorators.call_stack
    def slice(self, start: int = 0, end: int = None):
        """Returns the specified slice
        
        Args:
            start (int, optional): Start position. Defaults to 0.
            end (int, optional): End position. Defaults to None.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").slice(3, 6).o
            "e d"
        """
        self.state = self.state[start:end]
        return self

    @ChepyDecorators.call_stack
    def find_replace(self, pattern: str, repl: str, ignore_case=True):
        """Replace matched pattern with repln
        
        Args:
            pattern (str): Required. Pattern to search
            repl (str): Required. Pattern to match
            ignore_case (bool, optional): Case insensitive. Defaults to True.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some some data").find_replace(r"some\s", "data").o
            "datadatadata"
        """
        flags = 0
        if ignore_case:
            flags = re.IGNORECASE
        self.state = re.sub(pattern, repl, self._convert_to_str(), flags=flags)
        return self

    @ChepyDecorators.call_stack
    def escape_string(self):
        """Escape all special characters in a string
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.escape(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def unescape_string(self):
        """Unescape \\ from a string
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("\\$ome' d@ta").unescape_string().o
            "$ome' d@ta"
        """
        self.state = re.sub(r"\\", "", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def color_hex_to_rgb(self):
        """Convert hex color to rgb
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = tuple(
            int(self._convert_to_str().strip("#")[i : i + 2], 16) for i in (0, 2, 4)
        )
        return self

    @ChepyDecorators.call_stack
    def diff(
        self,
        state: int = None,
        buffer: int = None,
        colors: bool = False,
        swap: bool = False,
    ):
        """Diff state with another state or buffer
        
        Args:
            state (int, optional): Index of state to compare against. Defaults to None.
            buffer (int, optional): Index of buffer to compare against. Defaults to None.
            colors (int, optional): Show colored diff. Defaults to False.
            swap (int, optional): Swap the diff order. Defaults to False.
        
        Raises:
            TypeError: If both state and buffer is set to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("first string", "First $trin")
            >>> # there are two states
            >>> c.diff(state=1) # this will diff state 0 with state 1
            {F->f}irst {-$}strin{+g}
            >>> # optionally set colors=True in the diff method to see colored output
        """
        if state is not None and buffer is None:
            data = self.states.get(state)
            if isinstance(data, bytes):  # pragma: no cover
                data = data.decode()
        elif state is None and buffer is not None:
            data = self.buffers.get(buffer)
            if isinstance(data, bytes):  # pragma: no cover
                data = data.decode()
        else:  # pragma: no cover
            raise TypeError("Only select a state or a buffer to diff against")

        if swap:  # pragma: no cover
            matcher = difflib.SequenceMatcher(None, self._convert_to_str(), data)
        else:
            matcher = difflib.SequenceMatcher(None, data, self._convert_to_str())

        def process_tag(tag, i1, i2, j1, j2):  # pragma: no cover
            if tag == "replace":
                if colors:
                    return _int_colors.blue(matcher.b[j1:j2])
                else:
                    return "{" + matcher.a[i1:i2] + "->" + matcher.b[j1:j2] + "}"
            if tag == "delete":
                if colors:
                    return _int_colors.red(matcher.a[i1:i2])
                else:
                    return "{-" + matcher.a[i1:i2] + "}"
            if tag == "equal":
                return matcher.a[i1:i2]
            if tag == "insert":
                if colors:
                    return _int_colors.green(matcher.b[j1:j2])
                else:
                    return "{+" + matcher.b[j1:j2] + "}"
            assert False, "Unknown tag %r" % tag

        self.state = "".join(process_tag(*t) for t in matcher.get_opcodes())
        return self

    @ChepyDecorators.call_stack
    def pad(self, width: int, direction: str = "left", char: str = " "):
        """Pad string with a character
        
        Args:
            width (int): Required. Total length of string. The padding is calculated from 
                the length of state minus width.
            direction (str, optional): Padding direction. left or right. Defaults to 'left'.
            char (str, optional): Char to fill with. Defaults to ' '.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("lol").pad(5, char="a")
            lol # this is because "lol" is len 3, and padding is 5 - 2, which is 2, so no 
            padding is added
            >>> c = Chepy("lol").pad(8, char="a")
            lolaa # 8 - 3 so the output is padded for 5 chars
            >>> c = Chepy("lol").pad(8, direction="right", char="a")
            aalol
        """
        assert direction in [
            "left",
            "right",
        ], "Direction has to be either left or right"
        if direction == "left":
            self.state = self._convert_to_str().ljust(
                width - len(self._convert_to_str()), char
            )
        elif direction == "right":
            self.state = self._convert_to_str().rjust(
                width - len(self._convert_to_str()), char
            )
        return self

    @ChepyDecorators.call_stack
    def count(self):
        """Count anything
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = pydash.count_by(self.state)
        return self

    @ChepyDecorators.call_stack
    def set(self):
        """Get an array of unique values
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = list(set(self.state))
        return self

    @ChepyDecorators.call_stack
    def regex_to_str(self, all_combo: bool = False):
        """Convert a regex to a matching string
        
        Args:
            all_combo (bool, optional): Generate all combos that match regex. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 
        """
        if all_combo:
            self.state = list(exrex.generate(self._convert_to_str()))
        else:
            self.state = exrex.getone(self._convert_to_str())
        return self




from chepy.extras.characters import base64_char_sets


def test_base64_chars():
    assert base64_char_sets() == {
        "standard": "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",
        "url_safe": "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_",
        "filename_safe": "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+\\-=",
        "itoa64": "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz=",
        "xml": "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.",
        "y64": "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._-",
        "z64": "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+/=",
        "radix64": "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz+/=",
        "uuencoding": " -_",
        "xxencoding": "+\\-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
        "unix_crypt": "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
    }




import itertools
from typing import List, Any, Iterator


def generate_combo(
    words: List[Any], min_length: int = 0, max_length: int = None, join_by: str = None
) -> Iterator[tuple]:
    """Creates all possible combinations from the `words` being passed. 
    Returns a generator. `length` controls the length of the permutations.
    
    Args:
        words (List[Any]): List of strings.
        min_length (int, optional): Mininum length of permutations. By default, 0 which 
            will generate all
        max_length (int, optional): Maximum length of permutations. 
            By default, it is the length of the `words` list
    
    Returns:
        Iterator[tuple]: A generator containing tuples of combinations
    """
    if min_length > 0 and max_length is not None:
        for L in range(min_length, max_length + 1):
            for subset in itertools.permutations(words, L):
                yield join_by.join(subset) if join_by is not None else subset
    elif max_length is None:
        for L in range(min_length, len(words) + 1):
            for subset in itertools.permutations(words, L):
                yield join_by.join(subset) if join_by is not None else subset
    else:
        for subset in itertools.permutations(words, max_length):
            yield join_by.join(subset) if join_by is not None else subset


def hex_chars() -> list:
    """Returns an array of all the hex characters
    
    Returns:
        list: List of all hex characters
    """
    return list("{:02x}".format(x) for x in range(0, 256))




import logging

try:
    from markdown import markdown
except ImportError:
    logging.warning("Could not import markdown. Use pip install Markdown")

import chepy.core


class Chepy_Markdown(chepy.core.ChepyCore):
    """This plugin allows converting markdown to html
    """

    @chepy.core.ChepyDecorators.call_stack
    def markdown_to_html(self):
        """Convert markdown syntax to html
        
        Returns:
            ChepyPlugin: The Chepy object. 
        """
        self.state = markdown(self._convert_to_str())
        return self



import sys
import logging
import importlib
import inspect
import pkgutil
import json
from pathlib import Path
from configparser import ConfigParser


class ChepyConfig(object):
    def __init__(self):
        home = Path.home()
        self.chepy_dir = (
            Path(".chepy").resolve()
            if Path(".chepy").exists()
            else Path(home / ".chepy")
        )
        self.chepy_conf = Path(self.chepy_dir / "chepy.conf")

        if not self.chepy_conf.exists():  # pragma: no cover
            self.chepy_dir.mkdir(exist_ok=True)
            c = ConfigParser()

            c["Plugins"] = {
                "EnablePlugins": "false",
                "PluginPath": str(Path(__file__).parent / "chepy_plugins"),
            }
            c["Cli"] = {}
            cli_options = c["Cli"]
            cli_options["history_path"] = str(self.chepy_dir / "chepy_history")
            cli_options["prompt_char"] = ">"
            cli_options["prompt_colors"] = "#00ffff #ff0000 #ffd700"
            cli_options["show_rprompt"] = "false"
            cli_options["prompt_rprompt"] = "#00ff48"
            cli_options["prompt_bottom_toolbar"] = "#000000"
            cli_options["prompt_toolbar_version"] = "#00ff48"
            cli_options["prompt_toolbar_states"] = "#60cdd5"
            cli_options["prompt_toolbar_buffers"] = "#ff00ff"
            cli_options["prompt_toolbar_type"] = "#ffd700"
            cli_options["prompt_toolbar_plugins"] = "#ffccbc"
            cli_options["prompt_toolbar_errors"] = "#ff0000"
            cli_options["prompt_cli_method"] = "#ffd700"
            cli_options["prompt_plugin_method"] = "#30d8ff"
            cli_options["cli_info_color"] = "#c2c2ff"

            Path(str(self.chepy_dir / "chepy_history")).touch()
            if not self.chepy_conf.exists():
                with open(str(self.chepy_conf), "w") as f:
                    c.write(f)

        self.config = ConfigParser()
        self.config.read(str(self.chepy_conf))

        plugin_path = self.__get_conf_value("None", "PluginPath", "Plugins")
        self.enable_plugins = json.loads(
            self.__get_conf_value("false", "EnablePlugins", "Plugins")
        )

        if self.enable_plugins:
            if plugin_path != "None":
                self.plugin_path = Path(plugin_path).expanduser().resolve()
            else:
                self.plugin_path = Path(plugin_path)
        else:
            self.plugin_path = Path("None")

        self.history_path = self.__get_conf_value(
            str(self.chepy_dir / "chepy_history"), "history_path"
        )
        self.prompt_char = self.__get_conf_value(">", "prompt_char")
        self.prompt_colors = self.__get_conf_value(
            "#00ffff #ff0000 #ffd700", "prompt_colors"
        )
        self.show_rprompt = json.loads(self.__get_conf_value("false", "show_rprompt"))
        self.prompt_rprompt = self.__get_conf_value("#00ff48", "prompt_rprompt")
        self.prompt_bottom_toolbar = self.__get_conf_value(
            "#000000", "prompt_bottom_toolbar"
        )
        self.prompt_toolbar_version = self.__get_conf_value(
            "#00ff48", "prompt_toolbar_version"
        )
        self.prompt_toolbar_states = self.__get_conf_value(
            "#60cdd5", "prompt_toolbar_states"
        )
        self.prompt_toolbar_buffers = self.__get_conf_value(
            "#ff00ff", "prompt_toolbar_buffers"
        )
        self.prompt_toolbar_type = self.__get_conf_value(
            "#ffd700", "prompt_toolbar_type"
        )
        self.prompt_toolbar_plugins = self.__get_conf_value(
            "#ffccbc", "prompt_toolbar_plugins"
        )
        self.prompt_toolbar_errors = self.__get_conf_value(
            "#ff0000", "prompt_toolbar_errors"
        )
        self.prompt_cli_method = self.__get_conf_value(
            "#ffd700", "prompt_cli_method"
        )
        self.prompt_plugin_method = self.__get_conf_value(
            "#30d8ff", "prompt_plugin_method"
        )
        self.cli_info_color = self.__get_conf_value(
            "#c2c2ff", "cli_info_color"
        )

    def __get_conf_value(self, default: str, option: str, section: str = "Cli"):
        if self.config.has_section(section):
            if self.config.has_option(section, option):
                return self.config[section][option]
            else:
                return default
        else:
            return default

    def load_plugins(self):  # pragma: no cover
        plugins = []
        if self.plugin_path.stem != "None":
            sys.path.append(str(self.plugin_path))

            my_plugins = [
                importlib.import_module(name)
                for finder, name, ispkg in pkgutil.iter_modules()
                if (name.startswith("chepy_") and name != "chepy_plugins")
            ]

            for plugin in my_plugins:
                try:
                    klass, mod = inspect.getmembers(plugin, inspect.isclass)[0]
                    loaded = getattr(plugin, klass)
                    plugins.append(loaded)
                except:
                    logging.warning("Error loading {}".format(plugin.__name__))
        return plugins



# pylint: disable=undefined-variable
import sys
from setuptools import setup, find_packages
from os import path

# get version and author information
with open("chepy/__version__.py", "r") as f:
    exec(f.read())


def read_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


requirements = read_requirements()
# if sys.platform == "linux":
#     requirements.append("python-magic")
# else:
#     requirements.append("python-magic-bin==0.4.14")

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

core_extra_deps = ["requests"]

plugin_deps = [
    "scapy",
    "Markdown",
    "chepy",
    "pefile",
    "pyelftools",
    "ua-parser==0.8.0",
    "pydriller",
    "pyexiftool",
]

setup(
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="chepy",
    license="GPL",
    version=__version__,
    author=__author__,
    url="https://github.com/securisec/chepy",
    project_urls={
        "Documentation": "https://chepy.readthedocs.io/en/latest/",
        "Source Code": "https://github.com/securisec/chepy",
    },
    extras_require={"extras": core_extra_deps + plugin_deps},
    packages=find_packages(exclude=(["tests", "docs"])),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["chepy = chepy.__main__:main"]},
)



import emoji
import unicodedata
import regex as re

from ..core import ChepyCore, ChepyDecorators


class Language(ChepyCore):
    @ChepyDecorators.call_stack
    def unicode_chrs_by_lang(self, lang: str):
        """Detect language specific characters
        
        Detect characters from varios Unicode code point ids. Example 
        of languages are Common, Arabic, Armenian, Bengali, Bopomofo, Braille, 
        Buhid, Canadian_Aboriginal, Cherokee, Cyrillic, Devanagari, Ethiopic, 
        Georgian, Greek, Gujarati, Gurmukhi, Han, Hangul, Hanunoo, Hebrew, 
        Hiragana, Inherited, Kannada, Katakana, Khmer, Lao, Latin, Limbu, 
        Malayalam, Mongolian, Myanmar, Ogham, Oriya, Runic, Sinhala, Syriac, 
        Tagalog, Tagbanwa, TaiLe, Tamil, Telugu, Thaana, Thai, Tibetan, Yi, 
        but other code points should work also.
        
        Args:
            lang (str): Required. A string value identifying the language. 
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"\p{" + lang + "}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def find_emojis(self):
        """Find emojis, symbols, pictographs, map symbols and flags
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = emoji.get_emoji_regexp().findall(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def encode_utf_16_le(self):
        """Encode string as UTF16LE (1200). 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "utf_16_le", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_utf_16_le(self):
        """Decode string as UTF16LE (1200). 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("utf_16_le")
        return self

    @ChepyDecorators.call_stack
    def encode_utf_16_be(self):
        """Encode string as UTF16BE (1201). 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "utf_16_be", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_utf_16_be(self):
        """Decode string as UTF16BE (1201). 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("utf_16_be")
        return self

    @ChepyDecorators.call_stack
    def encode_utf_7(self):
        """Encode string as UTF7. 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("utf_7", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_utf_7(self):
        """Decode string as UTF7. 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("utf_7")
        return self

    @ChepyDecorators.call_stack
    def encode_cp500(self):
        """Encode string as EBCDIC-CP-BE, EBCDIC-CP-CH, IBM500 or CP500. 
        Western European languages. 
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp500", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp500(self):
        """Decode string as EBCDIC-CP-BE, EBCDIC-CP-CH, IBM500 or CP500. 
        Western European languages.
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp500")
        return self

    @ChepyDecorators.call_stack
    def encode_cp037(self):
        """Encode IBM037, IBM039. English languages.
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp037", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp037(self):
        """Decode IBM037, IBM039. English languages
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp037")
        return self

    @ChepyDecorators.call_stack
    def encode_cp874(self):
        """Encode Windows-874 Thai (874)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp874", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp874(self):
        """Decode Windows-874 Thai (874)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp874")
        return self

    @ChepyDecorators.call_stack
    def encode_cp932(self):
        """Encode Japanese Shift-JIS (932), 932, ms932, mskanji, ms-kanji	
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp932", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp932(self):
        """Decode Japanese Shift-JIS (932), 932, ms932, mskanji, ms-kanji	
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp932")
        return self

    @ChepyDecorators.call_stack
    def encode_gbk(self):
        """Encode Simplified Chinese GBK (936), 936, cp936, ms936
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("gbk", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_gbk(self):
        """Decode Simplified Chinese GBK (936), 936, cp936, ms936
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("gbk")
        return self

    @ChepyDecorators.call_stack
    def encode_gb2312(self):
        """Encode Simplified Chinese GB2312 (20936), chinese, csiso58gb231280, 
        euc-cn, euccn, eucgb2312-cn, gb2312-1980, gb2312-80, iso-ir-58
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("gb2312", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_gb2312(self):
        """Decode Simplified Chinese GB2312 (20936), chinese, csiso58gb231280, 
        euc-cn, euccn, eucgb2312-cn, gb2312-1980, gb2312-80, iso-ir-58
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("gb2312")
        return self

    @ChepyDecorators.call_stack
    def encode_cp949(self):
        """Encode Korean (949), 949, ms949, uhc
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp949", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp949(self):
        """Decode Korean (949), 949, ms949, uhc
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp949")
        return self

    @ChepyDecorators.call_stack
    def encode_cp950(self):
        """Encode Traditional Chinese Big5 (950), 950, ms950
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp950", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp950(self):
        """Decode Traditional Chinese Big5 (950), 950, ms950
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp950")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1250(self):
        """Encode Windows-1250 Central / Eastern European (1250), windows-1250	
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1250", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1250(self):
        """Decode Windows-1250 Central / Eastern European (1250), windows-1250	
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1250")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1251(self):
        """Encode Windows-1251 Cyrillic (1251), windows-1251
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1251", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1251(self):
        """Decode Windows-1251 Cyrillic (1251), windows-1251
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1251")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1252(self):
        """Encode Windows-1252 Latin (1252)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1252", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1252(self):
        """Decode Windows-1252 Latin (1252)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1252")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1253(self):
        """Encode Windows-1253 Greek (1253)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1253", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1253(self):
        """Decode Windows-1253 Greek (1253)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1253")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1254(self):
        """Encode Windows-1254 Turkish (1254)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1254", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1254(self):
        """Decode Windows-1254 Turkish (1254)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1254")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1255(self):
        """Encode Windows-1255 Hebrew (1255)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1255", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1255(self):
        """Decode Windows-1255 Hebrew (1255)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1255")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1256(self):
        """Encode Windows-1256 Arabic (1256)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1256", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1256(self):
        """Decode Windows-1256 Arabic (1256)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1256")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1257(self):
        """Encode Windows-1257 Baltic (1257)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1257", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1257(self):
        """Decode Windows-1257 Baltic (1257)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1257")
        return self

    @ChepyDecorators.call_stack
    def encode_cp1258(self):
        """Encode Windows-1258 Vietnam (1258)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode("cp1258", errors="backslashreplace")
        return self

    @ChepyDecorators.call_stack
    def decode_cp1258(self):
        """Decode Windows-1258 Vietnam (1258)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("cp1258")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_2(self):
        """Encode ISO-8859-2 Latin 2 Central European (28592)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_2", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_2(self):
        """Decode ISO-8859-2 Latin 2 Central European (28592)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_2")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_3(self):
        """Encode ISO-8859-3 Latin 3 South European (28593)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_3", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_3(self):
        """Decode ISO-8859-3 Latin 3 South European (28593)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_3")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_4(self):
        """Encode ISO-8859-4 Latin 4 North European (28594)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_4", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_4(self):
        """Decode ISO-8859-4 Latin 4 North European (28594)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_4")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_5(self):
        """Encode ISO-8859-5 Latin/Cyrillic (28595)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_5", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_5(self):
        """Decode ISO-8859-5 Latin/Cyrillic (28595)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_5")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_6(self):
        """Encode ISO-8859-6 Latin/Arabic (28596)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_6", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_6(self):
        """Decode ISO-8859-6 Latin/Arabic (28596)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_6")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_7(self):
        """Encode ISO-8859-7 Latin/Greek (28597)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_7", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_7(self):
        """Decode ISO-8859-7 Latin/Greek (28597)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_7")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_8(self):
        """Encode ISO-8859-8 Latin/Hebrew (28598)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_8", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_8(self):
        """Decode ISO-8859-8 Latin/Hebrew (28598)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_8")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_9(self):
        """Encode ISO-8859-9 Latin 5 Turkish (28599)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_9", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_9(self):
        """Decode ISO-8859-9 Latin 5 Turkish (28599)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_9")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_10(self):
        """Encode ISO-8859-10 Latin 6 Nordic (28600)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_10", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_10(self):
        """Decode ISO-8859-10 Latin 6 Nordic (28600)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_10")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_11(self):
        """Encode ISO-8859-11 Latin/Thai (28601)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_11", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_11(self):
        """Decode ISO-8859-11 Latin/Thai (28601)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_11")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_13(self):
        """Encode ISO-8859-13 Latin 7 Baltic Rim (28603)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_13", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_13(self):
        """Decode ISO-8859-13 Latin 7 Baltic Rim (28603)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_13")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_14(self):
        """Encode ISO-8859-14 Latin 8 Celtic (28604)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_14", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_14(self):
        """Decode ISO-8859-14 Latin 8 Celtic (28604)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_14")
        return self

    @ChepyDecorators.call_stack
    def encode_iso8859_15(self):
        """Encode ISO-8859-15 Latin 9 (28605)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_str().encode(
            "iso8859_15", errors="backslashreplace"
        )
        return self

    @ChepyDecorators.call_stack
    def decode_iso8859_15(self):
        """Decode ISO-8859-15 Latin 9 (28605)
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = self._convert_to_bytes().decode("iso8859_15")
        return self

    @ChepyDecorators.call_stack
    def remove_diacritics(self):
        """Replaces accented characters latin character equivalent.
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = unicodedata.normalize("NFKD", self._convert_to_str()).encode(
            "ascii", errors="ignore"
        )
        return self

    @ChepyDecorators.call_stack
    def unicode_to_str(self):
        """Escape any \\u characters to its proper unicode representation
        
        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().decode("unicode-escape")
        return self



import sys
import inspect
import fire
import regex as re
import argparse
import subprocess
from pathlib import Path
from docstring_parser import parse as _parse_doc
from prompt_toolkit.completion import (
    Completer,
    Completion,
    FuzzyCompleter,
    merge_completers,
)
from prompt_toolkit.validation import ValidationError, Validator
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession

from chepy import Chepy
from chepy.__version__ import __version__
import chepy.modules.internal.cli as chepy_cli
from chepy.modules.internal.colors import red, yellow, cyan, magenta, green
from chepy.config import ChepyConfig

config = ChepyConfig()
options = []
chepy = dir(Chepy)
fire_obj = None
errors = []

prompt_colors = config.prompt_colors.split()


def get_style():
    return Style.from_dict(
        {
            "completion-menu.completion.current": "bg:#00aaaa #000000",
            # "completion-menu.completion": "bg:#008888 #ffffff",
            "completion-menu.completion.fuzzymatch.outside": "fg:#00aaaa",
            "prompt1": "{} bold".format(prompt_colors[0]),
            "prompt2": "{} bold".format(prompt_colors[1]),
            "prompt3": "{} bold".format(prompt_colors[2]),
            "state_index": "#ffd700",
            "rprompt": "fg:{}".format(config.prompt_rprompt),
            "bottom-toolbar": config.prompt_bottom_toolbar,
            "prompt_toolbar_version": "bg:{}".format(config.prompt_toolbar_version),
            "prompt_toolbar_states": "bg:{}".format(config.prompt_toolbar_states),
            "prompt_toolbar_buffers": "bg:{}".format(config.prompt_toolbar_buffers),
            "prompt_toolbar_type": "bg:{}".format(config.prompt_toolbar_type),
            "prompt_toolbar_plugins": "bg:{}".format(config.prompt_toolbar_plugins),
            "prompt_toolbar_errors": "bg:{}".format(config.prompt_toolbar_errors),
        }
    )


def get_options():
    global errors
    options = dict()
    for method in chepy:
        try:
            attributes = getattr(Chepy, method)
            if not method.startswith("_") and not isinstance(attributes, property):
                args = inspect.getfullargspec(attributes).args
                parsed_doc = _parse_doc(attributes.__doc__)
                if len(args) == 1:
                    options[method] = {
                        "options": list(
                            map(lambda d: {"flag": d, "meta": ""}, args[1:])
                        ),
                        "meta": parsed_doc.short_description,
                        "returns": parsed_doc.returns.type_name,
                    }
                else:
                    options[method] = {
                        "options": list(
                            map(
                                lambda d: {
                                    "flag": d[1],
                                    "meta": parsed_doc.params[d[0]].description,
                                },
                                enumerate(args[1:]),
                            )
                        ),
                        "meta": parsed_doc.short_description,
                        "returns": parsed_doc.returns.type_name,
                    }
        except:
            e_type, e_msg, e_traceback = sys.exc_info()
            errors.append((e_type.__name__, "Error parsing options in:", method))
            continue
    return options


def prompt_message(fire_obj):
    elements = [
        ("class:prompt1", config.prompt_char),
        ("class:prompt2", config.prompt_char),
        ("class:prompt3", config.prompt_char),
    ]
    try:
        elements.append(
            (
                "class:state_index",
                str(fire_obj._current_index)
                + "/"
                + str(len(fire_obj.current_states) - 1)
                + " ",
            )
        )
    except AttributeError:
        pass
    elements.append(("class:name", " "))
    return elements


def bottom_toolbar(fire_obj):
    global errors
    if isinstance(fire_obj, Chepy):
        states = len(fire_obj.states) - 1 if fire_obj is not None else 0
        current_state = fire_obj._current_index if fire_obj is not None else 0
        buffers = len(fire_obj.buffers) if fire_obj is not None else 0
        return [
            ("class:prompt_toolbar_version", "Chepy: {} ".format(__version__)),
            (
                "class:prompt_toolbar_states",
                " States: {current}/{total} ".format(
                    current=current_state, total=states
                ),
            ),
            (
                "class:prompt_toolbar_buffers",
                " Buffers: {total} ".format(total=buffers),
            ),
            (
                "class:prompt_toolbar_type",
                " State: {} ".format(type(fire_obj.state).__name__),
            ),
            (
                "class:prompt_toolbar_plugins",
                " Plugins: {} ".format(config.enable_plugins),
            ),
            ("class:prompt_toolbar_errors", " Errors: {} ".format(len(errors))),
        ]


class CustomValidator(Validator):
    def validate(self, document):
        text = document.text.split()
        if re.search(r"^(!|#|\?)", document.text):
            pass
        elif len(text) > 1:
            if not text[-2].startswith("--"):
                if (
                    not re.search(r"\"|'", text[-1])
                    and not text[-1].startswith("--")
                    and text[-1] not in list(get_options().keys())
                ):
                    raise ValidationError(
                        cursor_position=1,
                        message="{text} is not a valid Chepy method".format(
                            text=text[-1]
                        ),
                    )


class CustomCompleter(Completer):
    def get_completions(self, document, complete_event):
        global options
        method_dict = get_options()
        word = document.get_word_before_cursor()

        methods = list(method_dict.items())

        selected = document.text.split()
        if len(selected) > 0:
            selected = selected[-1]
            if selected.startswith("cli_"):
                methods = options
            elif not selected.startswith("--"):
                current = method_dict.get(selected)
                if current is not None:
                    has_options = method_dict.get(selected)["options"]
                    if has_options is not None:
                        options = [
                            ("--{}".format(o["flag"]), {"meta": o["meta"]})
                            for o in has_options
                        ]
                        methods = options + methods
            else:
                methods = options

        for method_name, method_docs in methods:
            if method_name.startswith(word):
                meta = (
                    method_docs["meta"]
                    if isinstance(method_docs, dict) and method_docs.get("meta")
                    else ""
                )
                not_chepy_obj = ""
                if method_docs.get("returns"):
                    if method_docs["returns"] == None:
                        not_chepy_obj = "bg:{}".format(config.prompt_cli_method)
                    elif method_docs["returns"] == "ChepyPlugin":
                        not_chepy_obj = "bg:{}".format(config.prompt_plugin_method)
                yield Completion(
                    method_name,
                    start_position=-len(word),
                    display_meta=meta,
                    style=not_chepy_obj,
                )


def get_current_type(obj):
    if config.show_rprompt:
        if obj:
            return type(obj).__name__
        else:
            return "Type of current state"
    else:
        return None


def parse_args(args):
    parse = argparse.ArgumentParser()
    parse.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )
    parse.add_argument("-r", "--recipe", dest="recipe", help="Run a Chepy recipe and exit")
    parse.add_argument("data", nargs="*")
    return parse.parse_args(args)


def main():
    global fire_obj
    last_command = []

    args = parse_args(sys.argv[1:])
    args_data = args.data

    if args.recipe:
        print(Chepy(*args_data).load_recipe(args.recipe).o)
    else:
        args_data.append("-")

        history_file = config.history_path
        session = PromptSession(
            history=FileHistory(history_file),
            style=get_style(),
            wrap_lines=True,
            auto_suggest=AutoSuggestFromHistory(),
        )
        try:
            while True:
                prompt = session.prompt(
                    prompt_message(fire_obj=fire_obj),
                    bottom_toolbar=bottom_toolbar(fire_obj),
                    completer=FuzzyCompleter(
                        merge_completers([CustomCompleter(), chepy_cli.CliCompleter()])
                    ),
                    validator=CustomValidator(),
                    rprompt=get_current_type(fire_obj),
                )

                # check and output any commands that start with cli_
                if re.match(r"^\!", prompt):
                    print(magenta(subprocess.getoutput(re.sub(r"^\!\s?", "", prompt))))
                # check if line is a comment
                elif re.match(r"^#", prompt):
                    print(cyan(prompt))
                # get help for a method
                elif re.match(r"^\?", prompt):
                    _method_name = re.match(r"^\?(\s?)+([\w_]+)", prompt).group(2)
                    chepy_cli.get_doc(_method_name)
                # check if method called is a cli method
                elif re.search(r"^cli_.+", prompt):
                    cli_method = prompt.split()[0]
                    cli_args = re.search(r"--(\w+)\s([\w\W]+)", prompt)
                    if cli_method == "cli_show_errors":
                        getattr(chepy_cli, "cli_show_errors")(errors)
                    elif cli_method == "cli_plugin_path":
                        getattr(chepy_cli, "cli_plugin_path")(config)
                    elif cli_method == "cli_go_back":
                        args_data = args_data[: -len(last_command + ["-"])]
                        print(cyan("Go back: {}".format(last_command)))
                    elif cli_method == "cli_delete_history":
                        Path(config.history_path).unlink()
                    elif cli_args:
                        getattr(chepy_cli, cli_method)(
                            fire_obj, **{cli_args.group(1): cli_args.group(2)}
                        )
                    else:
                        getattr(chepy_cli, cli_method)(fire_obj)

                else:
                    for method in chepy:
                        if not method.startswith("_") and not isinstance(
                            getattr(Chepy, method), property
                        ):
                            fire.decorators._SetMetadata(
                                getattr(Chepy, method),
                                fire.decorators.ACCEPTS_POSITIONAL_ARGS,
                                False,
                            )
                    args_data += prompt.split()
                    if args_data[-1] != "-":
                        args_data.append("-")
                    try:
                        last_command = prompt.split() + ["-"]
                        fire_obj = fire.Fire(Chepy, command=args_data)
                    # handle required args for methods
                    except fire.core.FireExit:
                        args_data = args_data[: -len(last_command)]
                    except TypeError as e:
                        print(red(e.message))
                    except SystemExit:
                        sys.exit()
                    except:
                        # go back to last working arg
                        e_type, e_msg, e_traceback = sys.exc_info()
                        print(red(e_type.__name__), yellow(e_msg.__str__()))
                        args_data = args_data[: -len(last_command)]
                        continue
        except KeyboardInterrupt:
            print(green("\nOKBye"))
            sys.exit()



from chepy.extras.misc import *


def test_shannon_entropy():
    assert shannon_entropy("some text") == 2.725480556997868
    assert shannon_entropy("some text", unit="hartley") == 0.8204514002553331
    assert shannon_entropy("some text", unit="natural") == 1.8891591637540215


def test_IC():
    with open("tests/files/hello", "rb") as f:
        data = f.read()
    assert index_of_coincidence(data) == 0



import codecs
import string
import itertools
import base64
import binascii
import jwt
import pathlib
import ujson
import pycipher
import regex as re

from Crypto.Util.Padding import pad, unpad
from Crypto.Cipher import ARC4
from Crypto.Cipher import DES, DES3, AES
from Crypto.Cipher import Blowfish

from ..core import ChepyCore, ChepyDecorators
from .internal.constants import EncryptionConsts
from ..extras.combinatons import hex_chars


class EncryptionEncoding(ChepyCore):
    """This class handles most operations related to various encryption 
    related operations. This class inherits the ChepyCore class, and all the 
    methods are also available from the Chepy class
    
    Examples:
        >>> from chepy import Chepy
        or
        >>> from chepy.modules.encryptionencoding import EncryptionEncoding
    """

    def __check_mode(self, mode) -> None:
        assert mode in ["CBC", "OFB", "CTR", "ECB"], "Not a valid mode."

    def _convert_key(self, key, iv, hex_key, hex_iv):  # pragma: no cover
        if isinstance(key, str):
            key = key.encode()
        if hex_key:
            key = binascii.unhexlify(key)
        if isinstance(iv, str):
            iv = iv.encode()
        if hex_iv:
            iv = binascii.unhexlify(iv)
        else:
            iv = binascii.unhexlify(binascii.hexlify(iv))
        return key, iv

    @ChepyDecorators.call_stack
    def rotate(self, rotate_by: int):
        """Rotate string by provided number
        
        Args:
            rotate_by (int): Required. Number to rotate by
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            In this example, we will rotate by 20

            >>> Chepy("some data").rotate(20).output
            "migy xunu"
        """
        lc = string.ascii_lowercase
        uc = string.ascii_uppercase
        lookup = str.maketrans(
            lc + uc, lc[rotate_by:] + lc[:rotate_by] + uc[rotate_by:] + uc[:rotate_by]
        )
        self.state = self.state.translate(lookup)
        return self

    @ChepyDecorators.call_stack
    def rot_13(self):
        """ROT-13 encoding
        
        A simple caesar substitution cipher which rotates alphabet 
        characters by the specified amount (default 13).
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = codecs.encode(self._convert_to_str(), "rot_13")
        return self

    @ChepyDecorators.call_stack
    def rot_47(self):
        """ROT 47 encoding
        
        A slightly more complex variation of a caesar cipher, which includes 
        ASCII characters from 33 '!' to 126 '~'. Default rotation: 47.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some").rot_47().output
            "D@>6"
        """
        x = []
        for i in range(len(self.state)):
            j = ord(self.state[i])
            if j >= 33 and j <= 126:
                x.append(chr(33 + ((j + 14) % 94)))
            else:  # pragma: no cover
                x.append(self.state[i])
        self.state = "".join(x)
        return self

    @ChepyDecorators.call_stack
    def xor(self, key: str, key_type: str = "hex", ascii: bool = False):
        """XOR state with a key

        Valid key formats are utf, hex and base64. Simple XOR cipher is a type 
        of additive cipher based on logical operation xor, which operates according 
        to the following principles.

        (A * B) + (!A * !B)

        A  B  A XOR B
        0  0     0
        1  0     1
        0  1     1
        1  1     0

        The main advantage of xor chipher is that the encyption is reversible with t
        he same logical operation.
        
        Args:
            key (str): Required. The key to xor by
            key_type (str, optional): The key type. Valid values are hex, utf and base64. Defaults to "hex".
            ascii (bool, optional): If the input is in ascii format
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("secret").xor(key="secret", key_type="utf").to_hex()
            000000000000
        """
        assert key_type in [
            "utf",
            "hex",
            "base64",
        ], "Valid key types are hex, utf and base64"

        if isinstance(key, int):
            key = str(key)
        if key_type == "utf":
            key = binascii.hexlify(key.encode())
        elif key_type == "base64":
            key = binascii.hexlify(base64.b64decode(key.encode()))
        key = binascii.unhexlify(key)
        xor = bytearray(b"")
        if ascii:
            for char, key_val in zip(self._convert_to_str(), itertools.cycle(key)):
                xor.append(ord(char) ^ key_val)
        else:
            for char, key_val in zip(self._convert_to_bytes(), itertools.cycle(key)):
                xor.append(char ^ key_val)

        self.state = xor
        return self

    @ChepyDecorators.call_stack
    def xor_bruteforce(self, length: int = 100):
        """Brute force single byte xor

        For multibyte xor bruteforce, use chepy.extras.crypto_extras.xor_bruteforce_multi 
        function
        
        Args:
            length (int, optional): How to bytes to bruteforce. Defaults to 100.
        
        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("pf`qfw").xor_bruteforce()
            {'00': bytearray(b'pf`qfw'),
            '01': bytearray(b'qgapgv'),
            '02': bytearray(b'rdbsdu'),
            '03': bytearray(b'secret'), # here is our secret xored with the hex key 03
            '04': bytearray(b'tbdubs'),
            '05': bytearray(b'ucetcr'),
            ...}
            >>> c.get_by_key("03").bytearray_to_str()
            secret
            >>> c.xor("03").bytearray_to_str()
            pf`qfw
        """
        original = self.state
        found = {}
        keys = hex_chars()
        self.state = original[:length]
        for key in keys:
            self.xor(key)
            found[key] = self.state
            self.state = original[:length]
        self.state = found
        return self

    @ChepyDecorators.call_stack
    def jwt_decode(self):
        """Decode a JWT token. Does not verify
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = {
            "payload": jwt.decode(self._convert_to_str(), verify=False),
            "header": jwt.get_unverified_header(self._convert_to_str()),
        }
        return self

    @ChepyDecorators.call_stack
    def jwt_verify(self, secret: str, algorithm: list = ["HS256"]):
        """Verify JWT token
        
        Args:
            secret (str): Required. Secret key for token
            algorithm (list, optional): Array of valid algorithms. Defaults to ["HS256"]
        
        Returns:
            Chepy: The Chepy object. 
        """
        self.state = jwt.decode(
            self._convert_to_str(), key=secret, algorithms=algorithm
        )
        return self

    @ChepyDecorators.call_stack
    def jwt_sign(self, secret: str, algorithms: str = "HS256"):
        """Sign a json/dict object in JWT
        
        Args:
            secret (str): Required. Secret to sign with
            algorithms (str, optional): Signing algorithm. Defaults to "HS256".
        
        Returns:
            Chepy: The Chepy object. 
        """
        if isinstance(self.state, dict):
            data = self.state
        elif isinstance(self.state, str):
            data = ujson.loads(self.state)
        self.state = jwt.encode(data, key=secret, algorithm=algorithms)
        return self

    @ChepyDecorators.call_stack
    def jwt_bruteforce(
        self, wordlist: str, b64_encode: bool = False, algorithm: list = ["HS256"]
    ):
        """Brute force JWT token secret

        This method will use the provided wordlist to try and bruteforce the 
        verification.
        
        Args:
            wordlist (str): Required. Path to a wordlist
            b64_encode (bool, optional): Encoded the words in base64. Defaults to False.
            algorithm (list, optional): Array of valid algorithms. Defaults to ["HS256"].
        
        Returns:
            Chepy: The Chepy object. 
        """
        with open(pathlib.Path(wordlist).expanduser().absolute()) as words:
            for word in words:
                try:
                    word = word.strip()
                    if b64_encode:  # pragma: no cover
                        word = base64.b64encode(word)
                    j = jwt.decode(self._convert_to_str(), word, algorithms=algorithm)
                    self.state = {
                        "paylod": j,
                        "header": jwt.get_unverified_header(self._convert_to_str()),
                        "secret": word,
                    }
                    return self
                except jwt.InvalidSignatureError:
                    continue
            else:  # pragma: no cover
                return self

    @ChepyDecorators.call_stack
    def rc4_encrypt(self, key: str, hex_key: bool = False):
        """Encrypt raw state with RC4
        
        Args:
            key (str): Required. Secret key
            hex_key (bool, optional): If key is in hex. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some data").rc4_encrypt("736563726574", hex_key=True).o
            b"9e59bf79a2c0b7d253"
        """
        if hex_key:
            key = binascii.unhexlify(key)
        if isinstance(key, str):
            key = key.encode()
        cipher = ARC4.new(key)
        self.state = binascii.hexlify(cipher.encrypt(self._convert_to_bytes()))
        return self

    @ChepyDecorators.call_stack
    def rc4_decrypt(self, key: str, hex_key: bool = False):
        """Decrypt raw state with RC4
        
        Args:
            key (str): Required. Secret key
            hex_key (bool, optional): If key is in hex. Defaults to False.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("9e59bf79a2c0b7d253").hex_to_str().rc4_decrypt("secret").o
            b"some data"
        """
        if hex_key:
            key = binascii.unhexlify(key)
        if isinstance(key, str):
            key = key.encode()
        cipher = ARC4.new(key)
        self.state = cipher.decrypt(self._convert_to_bytes())
        return self

    @ChepyDecorators.call_stack
    def des_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Encrypt raw state with DES

        DES is a previously dominant algorithm for encryption, and was published 
        as an official U.S. Federal Information Processing Standard (FIPS). It is 
        now considered to be insecure due to its small key size. DES uses a key 
        length of 8 bytes (64 bits).<br>Triple DES uses a key length of 24 bytes. 
        You can generate a password-based key using one of the KDF operations. 
        The Initialization Vector should be 8 bytes long. If not entered, it will 
        default to 8 null bytes. Padding: In CBC and ECB mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some data").des_encrypt("70617373776f7264", hex_key=True).o
            b"1ee5cb52954b211d1acd6e79c598baac"

            To encrypt using a differnt mode

            >>> Chepy("some data").des_encrypt("password", mode="CTR").o
            b"0b7399049b0267d93d"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = DES.new(key, mode=DES.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = DES.new(key, mode=DES.MODE_ECB)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = DES.new(key, mode=DES.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES.new(key, mode=DES.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def des_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Decrypt raw state encrypted with DES. 

        DES is a previously dominant algorithm for encryption, and was published 
        as an official U.S. Federal Information Processing Standard (FIPS). It is 
        now considered to be insecure due to its small key size. DES uses a key 
        length of 8 bytes (64 bits).<br>Triple DES uses a key length of 24 bytes. 
        You can generate a password-based key using one of the KDF operations. 
        The Initialization Vector should be 8 bytes long. If not entered, it will 
        default to 8 null bytes. Padding: In CBC and ECB mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("1ee5cb52954b211d1acd6e79c598baac").hex_to_str().des_decrypt("password").o
            b"some data"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = DES.new(key, mode=DES.MODE_CBC, iv=iv)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB":
            cipher = DES.new(key, mode=DES.MODE_ECB)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CTR":
            cipher = DES.new(key, mode=DES.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES.new(key, mode=DES.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def triple_des_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Encrypt raw state with Triple DES

        Triple DES applies DES three times to each block to increase key size. Key: 
        Triple DES uses a key length of 24 bytes (192 bits).<br>DES uses a key length 
        of 8 bytes (64 bits).<br><br>You can generate a password-based key using one 
        of the KDF operations. IV: The Initialization Vector should be 8 bytes long. 
        If not entered, it will default to 8 null bytes. Padding: In CBC and ECB 
        mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some data").triple_des_encrypt("super secret password !!", mode="ECB").o
            b"f8b27a0d8c837edc8fb00ea85f502fb4"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = DES3.new(key, mode=DES3.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = DES3.new(key, mode=DES3.MODE_ECB)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = DES3.new(key, mode=DES3.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES3.new(key, mode=DES3.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def triple_des_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Decrypt raw state encrypted with DES. 

        Triple DES applies DES three times to each block to increase key size. Key: 
        Triple DES uses a key length of 24 bytes (192 bits).<br>DES uses a key length 
        of 8 bytes (64 bits).<br><br>You can generate a password-based key using one 
        of the KDF operations. IV: The Initialization Vector should be 8 bytes long. 
        If not entered, it will default to 8 null bytes. Padding: In CBC and ECB 
        mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("f8b27a0d8c837edce87dd13a1ab41f96")
            >>> c.hex_to_str()
            >>> c.triple_des_decrypt("super secret password !!")
            >>> c.o
            b"some data"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = DES3.new(key, mode=DES3.MODE_CBC, iv=iv)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB":
            cipher = DES3.new(key, mode=DES3.MODE_ECB)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CTR":
            cipher = DES3.new(key, mode=DES3.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = DES3.new(key, mode=DES3.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def aes_encrypt(
        self,
        key: str,
        iv: str = "00000000000000000000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Encrypt raw state with AES

        Advanced Encryption Standard (AES) is a U.S. Federal Information Processing 
        Standard (FIPS). It was selected after a 5-year process where 15 competing 
        designs were evaluated.
        
        key: The following algorithms will be used based on the size of the 
            16 bytes = AES-128
            24 bytes = AES-192
            32 bytes = AES-256
        
        You can generate a password-based key using one of the KDF operations. 
        IV: The Initialization Vector should be 16 bytes long. If not entered, it will 
        default to 16 null bytes. Padding: In CBC and ECB mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. 
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some data").aes_encrypt("secret password!", mode="ECB").o
            b"5fb8c186394fc399849b89d3b6605fa3"
        """

        assert mode in ["CBC", "OFB", "CTR", "ECB", "GCM"], "Not a valid mode."

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = AES.new(key, mode=AES.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 16))
            return self
        elif mode == "ECB":
            cipher = AES.new(key, mode=AES.MODE_ECB)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 16))
            return self
        elif mode == "CTR":
            cipher = AES.new(key, mode=AES.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "GCM":
            cipher = AES.new(
                key,
                mode=AES.MODE_GCM,
                nonce=binascii.unhexlify("00000000000000000000000000000000"),
            )
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = AES.new(key, mode=AES.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def aes_decrypt(
        self,
        key: str,
        iv: str = "00000000000000000000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Decrypt raw state encrypted with DES. 

        Advanced Encryption Standard (AES) is a U.S. Federal Information Processing 
        Standard (FIPS). It was selected after a 5-year process where 15 competing 
        designs were evaluated.<br><br><b>Key:</b> The following algorithms will 
        be used based on the size of the 
        
        key:
            16 bytes = AES-128
            24 bytes = AES-192
            32 bytes = AES-256
        
        You can generate a password-based key using one of the KDF operations. 
        IV: The Initialization Vector should be 16 bytes long. If not entered, it will 
        default to 16 null bytes. Padding: In CBC and ECB mode, PKCS#7 padding will be used.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. 
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("5fb8c186394fc399849b89d3b6605fa3")
            >>> c.hex_to_str()
            >>> c.aes_decrypt("7365637265742070617373776f726421", hex_key=True)
            >>> c.o
            b"some data"
        """

        assert mode in ["CBC", "OFB", "CTR", "ECB", "GCM"], "Not a valid mode."

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = AES.new(key, mode=AES.MODE_CBC, iv=iv)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 16)
            return self
        elif mode == "ECB":
            cipher = AES.new(key, mode=AES.MODE_ECB)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 16)
            return self
        elif mode == "CTR":
            cipher = AES.new(key, mode=AES.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "GCM":
            cipher = AES.new(
                key,
                mode=AES.MODE_GCM,
                nonce=binascii.unhexlify("00000000000000000000000000000000"),
            )
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = AES.new(key, mode=AES.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def blowfish_encrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Encrypt raw state with Blowfish

        Blowfish is a symmetric-key block cipher designed in 1993 by 
        Bruce Schneier and included in a large number of cipher suites 
        and encryption products. AES now receives more attention.
        IV: The Initialization Vector should be 8 bytes long.

        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. Defaults to '0000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> Chepy("some data").blowfish_encrypt("password", mode="ECB").o
            b"d9b0a79853f139603951bff96c3d0dd5"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CBC, iv=iv)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "ECB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_ECB)
            self.state = cipher.encrypt(pad(self._convert_to_bytes(), 8))
            return self
        elif mode == "CTR":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CTR, nonce=b"")
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_OFB, iv=iv)
            self.state = cipher.encrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def blowfish_decrypt(
        self,
        key: str,
        iv: str = "0000000000000000",
        mode: str = "CBC",
        hex_key: bool = False,
        hex_iv: bool = True,
    ):
        """Encrypt raw state with Blowfish

        Blowfish is a symmetric-key block cipher designed in 1993 by 
        Bruce Schneier and included in a large number of cipher suites 
        and encryption products. AES now receives more attention.
        IV: The Initialization Vector should be 8 bytes long.
        
        Args:
            key (str): Required. The secret key
            iv (str, optional): IV for certain modes only. 
                Defaults to '00000000000000000000000000000000'.
            mode (str, optional): Encryption mode. Defaults to 'CBC'.
            hex_key (bool, optional): If the secret key is a hex string. Defaults to False.
            hex_iv (bool, optional): If the IV is a hex string. Defaults to True.
        
        Returns:
            Chepy: The Chepy object. 

        Examples:
            >>> c = Chepy("d9b0a79853f13960fcee3cae16e27884")
            >>> c.hex_to_str()
            >>> c.blowfish_decrypt("password")
            >>> c.o
            b"some data"
        """

        self.__check_mode(mode)

        key, iv = self._convert_key(key, iv, hex_key, hex_iv)

        if mode == "CBC":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CBC, iv=iv)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "ECB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_ECB)
            self.state = unpad(cipher.decrypt(self._convert_to_bytes()), 8)
            return self
        elif mode == "CTR":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_CTR, nonce=b"")
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self
        elif mode == "OFB":
            cipher = Blowfish.new(key, mode=Blowfish.MODE_OFB, iv=iv)
            self.state = cipher.decrypt(self._convert_to_bytes())
            return self

    @ChepyDecorators.call_stack
    def vigenere_encode(self, key: str):
        """Encode with Vigenere ciper
        
        Args:
            key (str): Required. The secret key
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("secret").vigenere_encode("secret").o
            "KIEIIM"
        """
        self.state = pycipher.Vigenere(key=key).encipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def vigenere_decode(self, key: str):
        """Decode Vigenere ciper
        
        Args:
            key (str): Required. The secret key
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("KIEIIM").vigenere_decode("secret").o
            "SECRET"
        """
        self.state = pycipher.Vigenere(key=key).decipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def affine_encode(self, a: int = 1, b: int = 1):
        """Encode with Affine ciper
        
        Args:
            a (int, optional): Multiplier value. Defaults to 1
            b (int, optional): Additive value. Defaults to 1
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("secret").affine_encode().o
            "TFDSFU"
        """
        self.state = pycipher.Affine(a=a, b=b).encipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def affine_decode(self, a: int = 1, b: int = 1):
        """Decode Affine ciper
        
        Args:
            a (int, optional): Multiplier value. Defaults to 1
            b (int, optional): Additive value. Defaults to 1
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("TFDSFU").affine_decode().o
            "SECRET"
        """
        self.state = pycipher.Affine(a=a, b=b).decipher(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def atbash_encode(self):
        """Encode with Atbash ciper
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("secret").atbash_encode().o
            "HVXIVG"
        """
        self.state = pycipher.Atbash().encipher(self._convert_to_str(), keep_punct=True)
        return self

    @ChepyDecorators.call_stack
    def atbash_decode(self):
        """Decode Atbash ciper
        
        Returns:
            Chepy: The Chepy oject. 

        Examples:
            >>> Chepy("hvxivg").atbash_decode().o
            "SECRET"
        """
        self.state = pycipher.Atbash().decipher(self._convert_to_str(), keep_punct=True)
        return self

    @ChepyDecorators.call_stack
    def to_morse_code(
        self,
        dot: str = ".",
        dash: str = "-",
        letter_delim: str = " ",
        word_delim: str = "\n",
    ):
        """Encode string to morse code
        
        Args:
            dot (str, optional): The char for dot. Defaults to ".".
            dash (str, optional): The char for dash. Defaults to "-".
            letter_delim (str, optional): Letter delimiter. Defaults to " ".
            word_delim (str, optional): Word delimiter. Defaults to "\\n".
        
        Returns:
            Chepy: The Chepy object. 
        """
        encode = ""
        morse_code_dict = EncryptionConsts.MORSE_CODE_DICT
        for k, v in morse_code_dict.items():
            morse_code_dict[k] = v.replace(".", dot).replace("-", dash)
        for word in self._convert_to_str().split():
            for w in word:
                encode += morse_code_dict.get(w.upper()) + letter_delim
            encode += word_delim
        self.state = encode
        return self

    @ChepyDecorators.call_stack
    def from_morse_code(
        self,
        dot: str = ".",
        dash: str = "-",
        letter_delim: str = " ",
        word_delim: str = "\n",
    ):
        """Decode morse code
        
        Args:
            dot (str, optional): The char for dot. Defaults to ".".
            dash (str, optional): The char for dash. Defaults to "-".
            letter_delim (str, optional): Letter delimiter. Defaults to " ".
            word_delim (str, optional): Word delimiter. Defaults to "\\n".
        
        Returns:
            Chepy: The Chepy object. 
        """
        decode = ""
        morse_code_dict = EncryptionConsts.MORSE_CODE_DICT
        for k, v in morse_code_dict.items():
            morse_code_dict[k] = v.replace(".", dot).replace("-", dash)

        morse_code_dict = {value: key for key, value in morse_code_dict.items()}

        for chars in self._convert_to_str().split(letter_delim):
            if word_delim in chars:
                print("here", chars)
                chars = re.sub(word_delim, "", chars, re.I)
                print(chars)
                if morse_code_dict.get(chars) is not None:
                    decode += " " + morse_code_dict.get(chars)
            else:
                decode += morse_code_dict.get(chars)
        self.state = decode
        return self
